# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2022 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #

import datetime

import bs_audit

from bs_lib import query_builder
from bs_lib import Pod


class TableWatch:
    """
    This object sets up a watch on an interval of your choice. If table/s that
    you are watching change, you will be notified when changes have occured.
    """

    def __init__(self, pod: Pod, dao_name: str, watch_list: list = None, interval: int = 0):
        """
        Constructor.

        :param pod: The pod object to use.
        :param dao_name: Name of dao module to use.
        :param watch_list: The table watch list, if none, then all tables are watched.
        :param interval: How many seconds wait before checking for changes.
        """
        self._pod       = pod
        self._dao       = bs_audit.dao_by_name(dao_name)
        self.watch_list = watch_list or []
        self.interval   = interval
        self.last_check = datetime.datetime.min
        self.next_check = datetime.datetime.min


    def __del__(self):
        """
        Destructor.
        """
        pass


    def get_changed(self, watch_list: list = None, from_check_time: datetime.datetime = None) -> list:
        """
        Gets all the tables that have had changes since the last time
        that you made this method call.

        :param watch_list: Optionally give a custom watch list to check.
        :param from_check_time: Override the lastCheck time, and check from here instead.
        :return: A list of table id's that have changed, None if nothing has changed.
        """
        if from_check_time is None:
            dtnow = datetime.datetime.now()

            if dtnow < self.next_check:
                return []

            self.last_check = dtnow

            if self.interval > 0:
                self.next_check = dtnow + datetime.timedelta(seconds=self.interval)
            else:
                self.next_check = dtnow

        qry = self._dao.dCfgChangedSince(self._pod.dbcon)
        qry.irec.since = from_check_time or self.last_check

        if watch_list or self.watch_list:
            qry.irec.crit += query_builder(watch_list, 'c', 'id')

        res = []

        qry.exec()

        while qry.fetch():
            res.append(qry.orec.id)

        return res



# This file was generated by mettle.genes.db.GenPy3 [ver 2.2] on Thu Aug 11 12:14:36 2022
#

import datetime
import uuid
import time
import dataclasses
import typing
import mettle.io
import mettle.db

@dataclasses.dataclass(slots=True)
class iCfgChangedSince(mettle.io.ISerializable):

    since: typing.Optional[datetime.datetime] = dataclasses.field(default=None, compare=False)
    criteria: str = dataclasses.field(default='', compare=False)

    def clear(self):
        """
        Clears all member variables for this database record.
        """
        self.since = None
        self.criteria = ''

    def _name(self) -> str:
        """
        Name of the record.

        :return: The name.
        """
        return 'iCfgChangedSince'

    def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
        """
        Serialize record to a stream.

        :param _w:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _w.write_start(_oname)
        _w.write_datetime("since", self.since)
        _w.write_string("criteria", self.criteria)
        _w.write_end(_oname)

    def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
        """
        Deserialize record from a stream.

        :param _r:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _r.read_start(_oname)
        self.since = _r.read_datetime("since")
        self.criteria = _r.read_string("criteria")
        _r.read_end(_oname)

    def _copy_from(self, rec: "iCfgChangedSince"):
        """
        Copies the column from the rec into this record.

        :param rec: The source record.
        """
        self.since = rec.since
        self.criteria = rec.criteria

    @staticmethod
    def _cache_davs(dvc=None):
        """
        Cache the DAV into the dav cache (or create one) and return it.

        :param dvc: (mettle.lib.DavCache), target cache, if None a new one is created and returned.
        """
        if dvc is None:
            dvc = mettle.lib.DavCache()

        dvc.add_targ("id", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("since", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("criteria", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))

        return dvc

    def _get_davs(self, dvc=None):
        iCfgChangedSince._cache_davs(dvc)


    class List(list, mettle.io.ISerializable):
        """
        List Class
        """

        def _name(self) -> str:
            """
            Name of the list.

            :return: The name
            """
            return 'iCfgChangedSince.List'

        def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
            """
            Serialize the list to a stream.

            :param _w:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _w.write_start_list(_oname, len(self))

            for _rec in self:
                _rec._serialize(_w)

            _w.write_end(_oname)

        def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
            """
            Deserialize the list from a stream.

            :param _r:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _cnt = _r.read_start_list(_oname)

            while _cnt >= 1:
                _rec  = iCfgChangedSince()
                _cnt -= 1
                _rec._deserialize(_r)
                self.append(_rec)

            _r.read_end(_oname)

        def _get_davs(self, dvc=None):
            return iCfgChangedSince._cache_davs(dvc)




# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2022 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #

import datetime

import mettle.db

from bs_lib.transactionable import Transactionable

import bs_audit

from .db.tables.oaud_search import oAudSearch
from .base_audit            import BaseAudit


class Audit(BaseAudit, Transactionable):
    """
    Audit an object, has built in methods to help with auditing
    mettle generated table objects.
    """

    def __init__(self, dbcon: mettle.db.IConnect, dao_name: str):
        """
        Constructor.

        :param dbcon: Mettle db connection.
        :param dao_name: Name of dao module to use.
        """
        BaseAudit.__init__(self)

        self._dbcon   = dbcon
        self._dao     = bs_audit.dao_by_name(dao_name)


    def begin(self):
        """
        Overload the Transactionable method.
        """
        self._site_id = 0
        self._src     = None
        self._who     = None
        self._diffs.clear()


    def rollback(self):
        """
        Overload Transactionable method, undo all diff changes up to now.
        """
        self._diffs.clear()


    def commit(self) -> int:
        """
        Overload Transactionable method, saves all the audit diffs into transaction.

        :return: (int) The transaction id of the audit.
        """
        if not self._diffs:
            return 0

        if self._site_id < 1:
            raise Exception('Cannot commit audit, site id required.')

        if not self._who:
            raise Exception('Cannot commit audit, user required.')

        if not self._src:
            self._src = '<N/A>'

        total_diffs = 0

        for dobj in self._diffs:
            self._init_table_cfg(dobj['name'])
            total_diffs += self._build_diffs(dobj)

        if total_diffs < 1:
            return 0

        tran = self._dao.dTran(self._dbcon)
        aud  = self._dao.dAud(self._dbcon)
        col  = self._dao.dCol(self._dbcon)
        tbls = []

        tran.insert_deft(self._site_id, self._who, self._src)

        for dobj in self._diffs:
            diffs = dobj['diffs']

            if len(diffs) < 1:
                continue

            kdata = dobj.get('adict')

            if kdata is None:
                kdata = dobj.get('bdict')

            aud.insert_deft(
                tran.rec.id,
                dobj['dtype'],
                dobj['name'],
                dobj['pk']
            )

            if dobj['name'] not in tbls:
                tbls.append(dobj['name'])

            for dcol, dcvals in dobj['diffs'].items():
                col.insert_deft(aud.rec.id, dcol, dcvals[0], dcvals[1])

        if tbls:
            self._dao.dCfgUpdLastChange(self._dbcon).exec_deft("'%s'" % "','".join(tbls))

        self._diffs.clear()

        return tran.rec.id


    def search(self,
               site_id: int,
               dt_from: datetime.datetime = None,
               dt_to:   datetime.datetime = None,
               tname:   str  = None,
               pk:      str  = None,
               usr_id:  str  = None,
               actions: list = None,
               tran_id: int  = 0) -> oAudSearch.List:
        """
        Runs a search of audit history.

        :param site_id: The site id to use.
        :param dt_from: The date time from, req if tran_id = 0.
        :param dt_to: The date time to, req if tran_id = 0.
        :param tname: Table name to search for, req if tran_id = 0.
        :param pk: The primary key of the table to search on, req if tran_id = 0.
        :param usr_id: Optionally only search for this users changes in the db.
        :param actions: Optionally give a list of actions to look for, C, U, D
        :param tran_id: The transaction to search for, 0 = ignore.
        :return: The search audit result list.
        """
        qry = self._dao.dAudSearch(self._dbcon)
        res = oAudSearch.List()

        self._prep_search(qry.irec, site_id, dt_from, dt_to, tname, pk, usr_id, actions, tran_id)

        qry.exec().fetch_all(res)

        return res


    def _init_table_cfg(self, name: str):
        """
        Initialiaze the table config.

        :param name: The table name.
        """
        tobj = self._cfg.get(name)

        if tobj:
            return tobj

        tcfg = self._dao.dCfg(self._dbcon)

        if not tcfg.select_one_deft(name):
            raise Exception('Table [%s] not found in Bitsmiths-Audit table config.' % name)

        tobj = {
            'id'    : tcfg.rec.id,
            'pk'    : tcfg.rec.col_pk.split('|'),
            'ign'   : tcfg.rec.col_ignr.split('|'),
        }

        self._cfg[tcfg.rec.id] = tobj

        return tobj



# This file was generated by mettle.genes.db.GenPy3 [ver 2.2] on Thu Aug 11 12:14:36 2022
#  Target Database = postgresql
#
from .aud import dAud
from .aud_select_one import dAudSelectOne
from .aud_insert import dAudInsert
from .aud_search import dAudSearch
from .cfg import dCfg
from .cfg_select_one import dCfgSelectOne
from .cfg_insert import dCfgInsert
from .cfg_upd_last_change import dCfgUpdLastChange
from .cfg_changed_since import dCfgChangedSince
from .col import dCol
from .col_insert import dColInsert
from .tran import dTran
from .tran_select_one import dTranSelectOne
from .tran_insert import dTranInsert



# This file was generated by mettle.genes.db.GenPy3 [ver 2.2] on Thu Aug 11 12:14:36 2022
#  Target Database = postgresql
#
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_audit.db.tables.icfg_changed_since import iCfgChangedSince
from bs_audit.db.tables.ocfg_changed_since import oCfgChangedSince

class dCfgChangedSince:
    __slots__ = ("_dbcon", "_dbstmnt", "irec", "orec")

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = iCfgChangedSince()
        self.orec     = None

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        self._dbstmnt = None

    def exec_deft(self,
                  since: datetime.datetime,
                  criteria: str) -> "dCfgChangedSince":
        """
        Execute the query by setting all the inputs.

        :param since: datetime.datetime
        :param criteria: str
        :return: Self for convenience.
        """
        self.irec.since = since
        self.irec.criteria = criteria

        return self.exec()

    def exec(self, irec: iCfgChangedSince = None) -> "dCfgChangedSince":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = self._dbcon.statement("CfgChangedSince", self._dbcon.STMNT_TYPE_READ)

        self._dbstmnt.sql("""select
  c.id,
from
  audit.cfg c
where
  last_chg >= :since
  [criteria]""")

        self._dbstmnt.dynamic("[criteria]", self.irec.criteria)

        self._dbstmnt.bind_in("since", self.irec.since, datetime.datetime)

        self._dbstmnt.bind_out("id", str)

        self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self

    def fetch(self, must_exist: bool = False) -> bool:
        """
        Fetches the next row from the executed query.

        :param must_exist:
        :return: True if a row was fetched, else False.
        """
        if not self._dbcon.fetch(self._dbstmnt):
            if must_exist:
                raise mettle.lib.xMettle(f"Record not found: [table:Cfg, proc:ChangedSince, row:{self.irec}]", err_code=mettle.lib.xMettle.eCode.DBNotFound)
            return False

        self.orec = oCfgChangedSince()

        self.orec.id = self._dbstmnt.result[0]

        return True

    def fetch_iter(self) -> oCfgChangedSince:
        """
        Yield all the rows one at a time.

        :return: Iterator for output rows
        """
        while self.fetch():
            yield self.orec

    def fetch_all(self, out_list: oCfgChangedSince.List, clear_list: bool = True) -> int:
        """
        Fetch all the rows into the specified list and return the list count.

        :param out_list:
        :param clear_list:
        :return: Number of rows fetched.
        """
        if clear_list:
            out_list.clear()

        while self.fetch():
            out_list.append(self.orec)

        return len(out_list)




# This file was generated by mettle.genes.db.GenPy3 [ver 2.2] on Thu Aug 11 12:14:36 2022
#  Target Database = postgresql
#
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_audit.db.tables.aud_key import tAudKey
from bs_audit.db.tables.aud import tAud

class dAudSelectOne:
    __slots__ = ("_dbcon", "_dbstmnt", "irec", "orec")

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tAudKey()
        self.orec     = None

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        self._dbstmnt = None

    def exec_deft(self,
                  id: int) -> "dAudSelectOne":
        """
        Execute the query by setting all the inputs.

        :param id: int
        :return: Self for convenience.
        """
        self.irec.id = id

        return self.exec()

    def exec(self, irec: tAudKey = None) -> "dAudSelectOne":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = self._dbcon.statement("AudSelectOne", self._dbcon.STMNT_TYPE_READ)

        self._dbstmnt.sql("""select
  id,
  tran_id,
  action,
  tbl_id,
  tbl_key
from
  audit.Aud
where
  id = :id""")

        self._dbstmnt.bind_in("id", self.irec.id, int, 8)

        self._dbstmnt.bind_out("id", int, 8)
        self._dbstmnt.bind_out("tran_id", int, 8)
        self._dbstmnt.bind_out("action", str, 1)
        self._dbstmnt.bind_out("tbl_id", str)
        self._dbstmnt.bind_out("tbl_key", str)

        self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self

    def fetch(self, must_exist: bool = False) -> bool:
        """
        Fetches the next row from the executed query.

        :param must_exist:
        :return: True if a row was fetched, else False.
        """
        if not self._dbcon.fetch(self._dbstmnt):
            if must_exist:
                raise mettle.lib.xMettle(f"Record not found: [table:Aud, proc:SelectOne, row:{self.irec}]", err_code=mettle.lib.xMettle.eCode.DBNotFound)
            return False

        self.orec = tAud()

        self.orec.id = self._dbstmnt.result[0]
        self.orec.tran_id = self._dbstmnt.result[1]
        self.orec.action = self._dbstmnt.result[2]
        self.orec.tbl_id = self._dbstmnt.result[3]
        self.orec.tbl_key = self._dbstmnt.result[4]

        return True

    def fetch_iter(self) -> tAud:
        """
        Yield all the rows one at a time.

        :return: Iterator for output rows
        """
        while self.fetch():
            yield self.orec

    def fetch_all(self, out_list: tAud.List, clear_list: bool = True) -> int:
        """
        Fetch all the rows into the specified list and return the list count.

        :param out_list:
        :param clear_list:
        :return: Number of rows fetched.
        """
        if clear_list:
            out_list.clear()

        while self.fetch():
            out_list.append(self.orec)

        return len(out_list)




# This file was generated by mettle.genes.db.GenPy3 [ver 2.2] on Thu Aug 11 12:14:36 2022
#  Target Database = postgresql
#
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_audit.db.tables.cfg import tCfg

class dCfgInsert:
    __slots__ = ("_dbcon", "_dbstmnt", "irec")

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tCfg()

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        self._dbstmnt = None

    def exec_deft(self,
                  id: str,
                  col_pk: str,
                  col_ignr: str,
                  par_id: str,
                  par_col: str,
                  last_chg: datetime.datetime,
                  audby_col: str,
                  mode: str) -> "dCfgInsert":
        """
        Execute the query by setting all the inputs.

        :param id: str
        :param col_pk: str
        :param col_ignr: str
        :param par_id: str
        :param par_col: str
        :param last_chg: datetime.datetime
        :param audby_col: str
        :param mode: str
        :return: Self for convenience.
        """
        self.irec.id = id
        self.irec.col_pk = col_pk
        self.irec.col_ignr = col_ignr
        self.irec.par_id = par_id
        self.irec.par_col = par_col
        self.irec.last_chg = last_chg
        self.irec.audby_col = audby_col
        self.irec.mode = mode

        return self.exec()

    def exec(self, irec: tCfg = None) -> "dCfgInsert":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = self._dbcon.statement("CfgInsert", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""insert into audit.Cfg (
  id,
  col_pk,
  col_ignr,
  par_id,
  par_col,
  last_chg,
  audby_col,
  mode
) values (
  :id,
  :col_pk,
  :col_ignr,
  :par_id,
  :par_col,
  :last_chg,
  :audby_col,
  :mode
)""")

        self._dbstmnt.bind_in("id", self.irec.id, str)
        self._dbstmnt.bind_in("col_pk", self.irec.col_pk, str)
        self._dbstmnt.bind_in("col_ignr", self.irec.col_ignr, str)
        self._dbstmnt.bind_in("par_id", self.irec.par_id, str)
        self._dbstmnt.bind_in("par_col", self.irec.par_col, str)
        self._dbstmnt.bind_in("last_chg", self.irec.last_chg, datetime.datetime)
        self._dbstmnt.bind_in("audby_col", self.irec.audby_col, str)
        self._dbstmnt.bind_in("mode", self.irec.mode, str, 1)

        self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self




# This file was generated by mettle.genes.db.GenPy3 [ver 2.2] on Thu Aug 11 12:14:36 2022
#  Target Database = postgresql
#
import asyncio
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_audit.db.tables.cfg import tCfg
from bs_audit.db.tables.cfg_key import tCfgKey

from .cfg_select_one import dCfgSelectOne
from .cfg_insert import dCfgInsert

class dCfg:
    __slots__ = ("_dbcon", "rec")

    def __init__(self, dbcon: mettle.db.IAConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self.rec    = tCfg()
        self._dbcon = dbcon

    def __enter__(self):
        """
        DAO enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        DAO exit.
        """
        pass

    async def try_select_one(self) -> bool:
        """
        Attempt to select the record.

        :return: True if the record was found.
        """
        _key = tCfgKey()

        self.rec._write_key(_key)

        return await self.try_select_one_by_key(_key)

    async def try_select_one_with(self, _rec: tCfg) -> bool:
        """
        Attempt to select the record using a table rec.

        :param _rec: The table rec to select with and into.
        :return: True if the row as selectd.
        """
        _key = tCfgKey()
        _rec._write_key(_key)

        if not await self.try_select_one_by_key(_key):
            return False

        _rec._copy_from(self.rec)

        return True

    async def try_select_one_deft(self,
                                  id: str) -> bool:
        """
        Attempt to select the record by it's primary key columns.

        :param id: str
        :return: True if the row as selectd.
        """
        _key = tCfgKey()

        _key.id = id

        return await self.try_select_one_by_key(_key)

    async def try_select_one_by_key(self, _key: tCfgKey) -> bool:
        """
        Attempt to select the record with it's primary key.

        :param _key: Primary key of the table record.
        :returns: True if row was selected.
        """
        with dCfgSelectOne(self._dbcon) as _qry:
            await _qry.exec(_key)
            if not await _qry.fetch():
                return False

            self.rec = _qry.orec

        return True

    async def select_one(self) -> "dCfg":
        """
        Selects the record, raises and error if it does not exist.

        :return: Self for convenience.
        """
        if not await self.try_select_one():
            raise mettle.lib.xMettle(f"Not found: [table:dCfg, row:{self.rec}]", err_code=mettle.lib.xMettle.eCode.DBNotFound)

        return self

    async def select_one_with(self, _rec: tCfg) -> "dCfg":
        """
        Attempt to select the record with the table record.

        :param _rec: Table record.
        :returns: Self for convenience.
        """
        if not await self.try_select_one_with(_rec):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dCfg)")

        return self

    async def select_one_deft(self,
                              id: str) -> "dCfg":
        """
        Selects the record by its primary key columns, raises and error if it does not exist.

        :param id: str
        """
        _key = tCfgKey()

        _key.id = id

        if not await self.try_select_one_by_key(_key):
            raise mettle.lib.xMettle(f"Not found: [table:dCfg, row:{_key}]", err_code=mettle.lib.xMettle.eCode.DBNotFound)

        return self

    async def select_one_by_key(self, _key: tCfgKey) -> "dCfg":
        """
        Attempt to select the record with it's primary key.

        :param _key: Primary key of the table record.
        :return: Self for convenience.
        """
        if not await self.try_select_one_by_key(_key):
            raise mettle.lib.xMettle(f"Not found: [table:dCfg, row:{self.irec}]", xMettle.eCode.DBNotFound)

        return self

    async def insert_deft(self,
                          id: str,
                          col_pk: str,
                          col_ignr: str,
                          par_id: str,
                          par_col: str,
                          last_chg: datetime.datetime,
                          audby_col: str,
                          mode: str):
        """
        Insert the record setting all column values.

        :param id: str
        :param col_pk: str
        :param col_ignr: str
        :param par_id: str
        :param par_col: str
        :param last_chg: datetime.datetime
        :param audby_col: str
        :param mode: str
        """
        self.rec.id = id
        self.rec.col_pk = col_pk
        self.rec.col_ignr = col_ignr
        self.rec.par_id = par_id
        self.rec.par_col = par_col
        self.rec.last_chg = last_chg
        self.rec.audby_col = audby_col
        self.rec.mode = mode

        await self.insert()

    async def insert(self, irec: tCfg = None):
        """
        Insert the record in the database with the record or irec if it is not None.

        :param irec: If not None use this record to insert into the database.
        """
        if irec:
            self.rec._copy_from(irec)

        with dCfgInsert(self._dbcon) as _qry:
            await _qry.exec(self.rec)

        if irec:
            irec._copy_from(self.rec)




# This file was generated by mettle.genes.db.GenPy3 [ver 2.2] on Thu Aug 11 12:14:36 2022
#  Target Database = postgresql
#
import asyncio
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_audit.db.tables.col import tCol
from bs_audit.db.tables.col_key import tColKey

from .col_insert import dColInsert

class dCol:
    __slots__ = ("_dbcon", "rec")

    def __init__(self, dbcon: mettle.db.IAConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self.rec    = tCol()
        self._dbcon = dbcon

    def __enter__(self):
        """
        DAO enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        DAO exit.
        """
        pass

    async def insert_deft(self,
                          aud_id: int,
                          name: str,
                          before: str,
                          after: str):
        """
        Insert the record setting all column values.

        :param aud_id: int
        :param name: str
        :param before: str
        :param after: str
        """
        self.rec.aud_id = aud_id
        self.rec.name = name
        self.rec.before = before
        self.rec.after = after

        await self.insert()

    async def insert(self, irec: tCol = None):
        """
        Insert the record in the database with the record or irec if it is not None.

        :param irec: If not None use this record to insert into the database.
        """
        if irec:
            self.rec._copy_from(irec)

        with dColInsert(self._dbcon) as _qry:
            await _qry.exec(self.rec)

        if irec:
            irec._copy_from(self.rec)




# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2022 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #

import datetime
import json
import uuid

import mettle.db
import mettle.io

from bs_lib     import query_builder

from .db.tables import iAudSearch


class BaseAudit:
    """
    Base audit class that has common functions
    """

    def __init__(self):
        """
        Constructor.
        """
        self._dbcon   = None
        self._dao     = None
        self._cfg     = {}
        self._diffs   = []
        self._site_id = 0
        self._src     = None
        self._who     = None


    def aud(self, site_id: int, src: str, who: str) -> "BaseAudit":
        """
        Set the source and user who is doing the changes.

        :param site_id: The site the audit is occuring for.
        :param src: The program/operation that is doing the audit.
        :param who: The user making the changes.
        :return: Returns itself for convenience.
        """
        self._site_id = site_id
        self._src    = src
        self._who    = who

        return self


    def diff(self, before: object, after: object, tbl_name: str = None) -> None:
        """
        Add a diff audit.

        :param before: The before object.
        :param after: The after object.
        :param tbl_name: Optionally specifiy the table name of before/after object
        """
        bname, bdict = self._read_object(before)
        aname, adict = self._read_object(after)

        if tbl_name is None and bname is None and aname is None:
            raise Exception('Table name not specified.')

        if not tbl_name:
            if bname and aname and bname != aname:
                raise Exception('Table names do not match [before:%s, after:%s]' % (bname, aname))

            tbl_name = bname or aname

        if not bdict and not adict:
            return

        self._diffs.append({
            'name'  : tbl_name,
            'bdict' : bdict,
            'adict' : adict,
            'diffs' : {},
            'aud'   : None,
        })


    def diff_bulk(self, before_list: list, after_list: list, tbl_name: str = None) -> None:
        """
        Bulk audits a set of items.

        :param before_list: List of before objects, can be None for bulk inserts.
        :param after_list: List of after objects, can be None for bulk delets.
        :param tbl_name: Optionally specifiy the table name of before/after objects.
        """
        if not before_list and not after_list:
            return

        if not before_list:
            for x in after_list:
                self.diff(None, x, tbl_name)
        elif not after_list:
            for x in before_list:
                self.diff(x, None, tbl_name)
        else:
            if len(before_list) != len(after_list):
                raise Exception('Before list [%d] and after list [%d] are not the same length!' % (
                    len(before_list), len(after_list)))

            idx = 0

            while idx < len(before_list):
                self.diff(before_list[idx], after_list[idx])
                idx += 1


    def _prep_search(self,
                     irec:    iAudSearch,
                     site_id: int,
                     dt_from: datetime.datetime,
                     dt_to:   datetime.datetime,
                     tname:   str,
                     pk:      str,
                     usr_id:  str,
                     actions: list,
                     tran_id: int):
        """
        Prepare the search query.

        :param site_id: The site id to use.
        :param dt_from: The date time from, req if tran_id = 0.
        :param dt_to: The date time to, req if tran_id = 0.
        :param tname: Table name to search for, req if tran_id = 0.
        :param pk: The primary key of the table to search on, req if tran_id = 0.
        :param usr_id: Optionally only search for this users changes in the db.
        :param actions: Optionally give a list of actions to look for, C, U, D
        :param tran_id: The transaction to search for, 0 = ignore.
        :return: The search audit result list.
        """
        irec.site_id = site_id

        if tran_id > 0:
            irec.date_from = None
            irec.date_fo   = None
            irec.criteria = query_builder.dyn_crit(tran_id, 't', 'id')
            return

        irec.date_from = dt_from
        irec.date_fo   = dt_to

        if usr_id:
            irec.criteria += query_builder.dyn_crit(usr_id, 't', 'who')

        if tname:
            irec.criteria += query_builder.dyn_crit(tname, 'a', 'tbl_id')

            if pk:
                irec.criteria += query_builder.dyn_crit(pk, 'a', 'tbl_key')

        if actions:
            irec.criteria += query_builder.dyn_list(actions, 'a', 'action')


    def _read_object(self, obj: object) -> tuple:
        """
        Attempts to read an object putting all its fields into a dictionary.

        :param obj: A class instance or dictionary.
        :return: (name, dict), note return values can be None depending on obj.
        """
        if obj is None:
            return (None, None)

        if type(obj) == dict:
            return (None, obj)

        name = None

        if isinstance(obj, mettle.io.ISerializable):
            name = obj._name()

        return (name, obj.__dict__)


    def _get_key(self, tname: str, key_type: str, cols: list, data: dict) -> str:
        """
        Gets the cols from the data.

        :param tname: Table name
        :param key_type: The key type.
        :param cols: List of key columns.
        :param data: All data columns.
        :return: Pipe delimetered cols string.
        """
        if len(cols) < 1:
            return ''

        res = ''

        for pcol in cols:
            if pcol == '':
                continue

            if pcol not in data:
                raise Exception('%s key column [%s] not found [table:%s]' % (
                    key_type, pcol, tname))

            val = data[pcol]

            if res == '':
                res = self._text_value(val, pcol)
            else:
                res += '|%s' % self._text_value(val, pcol)

        return res


    def _build_diffs(self, dobj: dict) -> int:
        """
        Work out the differenes for each dict object.

        :param dobj: Diff dictionary object.
        :return: Total diffs detected.
        """
        tobj  = self._cfg[dobj['name']]
        diff  = dobj['diffs']
        bdata = dobj['bdict']
        adata = dobj['adict']

        if bdata is None:
            dobj['dtype'] = 'C'

            for key, val in adata.items():
                if key in tobj['ign']:
                    continue

                diff[key] = ['', self._text_value(val, key)]

            for pk in tobj['pk']:
                if pk not in adata.keys():
                    raise Exception('Primary key [%s] not found for table [%s] - data[%s]' % (
                        pk, dobj['name'], str(adata)))

            dobj['pk'] = self._get_key(dobj['name'], 'Primary', tobj['pk'], adata)

        elif adata is None:
            dobj['dtype'] = 'D'

            for key, val in bdata.items():
                if key in tobj['ign']:
                    continue

                diff[key] = [self._text_value(val, key), '']

            dobj['pk'] = self._get_key(dobj['name'], 'Primary', tobj['pk'], bdata)

        else:
            dobj['dtype'] = 'U'
            read  = []

            for key, val in bdata.items():
                read.append(key)

                if key in tobj['ign']:
                    continue

                btxt = self._text_value(val, key)
                atxt = self._text_value(adata.get(key), key)

                if btxt != atxt:
                    diff[key] = [btxt, atxt]

            for key, val in adata.items():
                if key in read:
                    continue

                if key in tobj['ign']:
                    continue

                btxt = self._text_value(None, key)
                atxt = self._text_value(val, key)

                if btxt != atxt:
                    diff[key] = [btxt, atxt]

            bpk = self._get_key(dobj['name'], 'Primary', tobj['pk'], bdata)
            apk = self._get_key(dobj['name'], 'Primary', tobj['pk'], adata)

            if apk != bpk:
                raise Exception('Primary keys do not match [beforePK:%s, afterPK:%s, table:%s]' % (
                    bpk, apk, dobj['name']))

            dobj['pk'] = bpk

        return len(diff)


    def _text_value(self, val: object, pcol: str) -> str:
        """
        Gets the standard text value of the value.

        :param val: The value object.
        :param pcol: The column name.
        :return:  The text value.
        """
        if val is None:
            return ''

        if type(val) == str:
            return val

        if type(val) == int or type(val) == float or type(val) == bool:
            return str(val)

        if type(val) == datetime.date:
            if val == datetime.date.min:
                return ''

            return val.strftime('%Y-%m-%d')

        if type(val) == datetime.datetime:
            if val == datetime.datetime.min:
                return ''

            return val.strftime('%Y-%m-%d %H:%M:%S')

        if type(val) == datetime.time:
            return val.strftime('%H:%M:%S')

        if isinstance(val, dict) or isinstance(val, list):
            return json.dumps(val)

        if isinstance(val, uuid.UUID):
            return str(val)

        if type(val) == bytes or type(val) == bytearray:
            return ''

        raise Exception('Type [%s] not expected for audit purposes [column:%s].' % (
            str(type(val)), pcol))



# This file was generated by mettle.genes.db.GenPy3 [ver 2.2] on Thu Aug 11 12:14:36 2022
#  Target Database = postgresql
#
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_audit.db.tables.col import tCol

class dColInsert:
    __slots__ = ("_dbcon", "_dbstmnt", "irec")

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tCol()

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        self._dbstmnt = None

    def exec_deft(self,
                  aud_id: int,
                  name: str,
                  before: str,
                  after: str) -> "dColInsert":
        """
        Execute the query by setting all the inputs.

        :param aud_id: int
        :param name: str
        :param before: str
        :param after: str
        :return: Self for convenience.
        """
        self.irec.aud_id = aud_id
        self.irec.name = name
        self.irec.before = before
        self.irec.after = after

        return self.exec()

    def exec(self, irec: tCol = None) -> "dColInsert":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = self._dbcon.statement("ColInsert", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""insert into audit.Col (
  aud_id,
  name,
  before,
  after
) values (
  :aud_id,
  :name,
  :before,
  :after
)""")

        self._dbstmnt.bind_in("aud_id", self.irec.aud_id, int, 8)
        self._dbstmnt.bind_in("name", self.irec.name, str)
        self._dbstmnt.bind_in("before", self.irec.before, str)
        self._dbstmnt.bind_in("after", self.irec.after, str)

        self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self




# This file was generated by mettle.genes.db.GenPy3 [ver 2.2] on Thu Aug 11 12:14:36 2022
#

from .aud_key import tAudKey
from .aud import tAud
from .iaud_search import iAudSearch
from .oaud_search import oAudSearch
from .cfg_key import tCfgKey
from .cfg import tCfg
from .icfg_upd_last_change import iCfgUpdLastChange
from .icfg_changed_since import iCfgChangedSince
from .ocfg_changed_since import oCfgChangedSince
from .col_key import tColKey
from .col import tCol
from .tran_key import tTranKey
from .tran import tTran



# This file was generated by mettle.genes.db.GenPy3 [ver 2.2] on Thu Aug 11 12:14:36 2022
#  Target Database = postgresql
#
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_audit.db.tables.aud import tAud

class dAudInsert:
    __slots__ = ("_dbcon", "_dbstmnt", "irec")

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tAud()

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        self._dbstmnt = None

    def exec_deft(self,
                  id: int,
                  tran_id: int,
                  action: str,
                  tbl_id: str,
                  tbl_key: str) -> "dAudInsert":
        """
        Execute the query by setting all the inputs.

        :param id: int
        :param tran_id: int
        :param action: str
        :param tbl_id: str
        :param tbl_key: str
        :return: Self for convenience.
        """
        self.irec.id = id
        self.irec.tran_id = tran_id
        self.irec.action = action
        self.irec.tbl_id = tbl_id
        self.irec.tbl_key = tbl_key

        return self.exec()

    def exec(self, irec: tAud = None) -> "dAudInsert":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = self._dbcon.statement("AudInsert", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""insert into audit.Aud (
  tran_id,
  action,
  tbl_id,
  tbl_key
) values (
  :tran_id,
  :action,
  :tbl_id,
  :tbl_key
)""")

        self._dbstmnt.bind_in("tran_id", self.irec.tran_id, int, 8)
        self._dbstmnt.bind_in("action", self.irec.action, str, 1)
        self._dbstmnt.bind_in("tbl_id", self.irec.tbl_id, str)
        self._dbstmnt.bind_in("tbl_key", self.irec.tbl_key, str)

        self._dbcon.execute(self._dbstmnt)

        self.irec.id = self._dbcon.post_get_sequence(self._dbstmnt, "audit.Aud", "id", 8)

        if irec:
            irec._copy_from(self.irec)

        return self




# This file was generated by mettle.genes.db.GenPy3 [ver 2.2] on Thu Aug 11 12:14:36 2022
#  Target Database = postgresql
#
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_audit.db.tables.tran import tTran

class dTranInsert:
    __slots__ = ("_dbcon", "_dbstmnt", "irec")

    def __init__(self, dbcon: mettle.db.IAConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tTran()

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        self._dbstmnt = None

    async def exec_deft(self,
                        id: int,
                        site_id: int,
                        who: str,
                        src: str,
                        tm_stamp: datetime.datetime) -> "dTranInsert":
        """
        Execute the query by setting all the inputs.

        :param id: int
        :param site_id: int
        :param who: str
        :param src: str
        :param tm_stamp: datetime.datetime
        :return: Self for convenience.
        """
        self.irec.id = id
        self.irec.site_id = site_id
        self.irec.who = who
        self.irec.src = src
        self.irec.tm_stamp = tm_stamp

        return await self.exec()

    async def exec(self, irec: tTran = None) -> "dTranInsert":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = await self._dbcon.statement("TranInsert", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""insert into audit.Tran (
  site_id,
  who,
  src,
  tm_stamp
) values (
  :site_id,
  :who,
  :src,
  :tm_stamp
)""")

        self.irec.tm_stamp = await self._dbcon.get_timestamp()

        self._dbstmnt.bind_in("site_id", self.irec.site_id, int, 4)
        self._dbstmnt.bind_in("who", self.irec.who, str, 128)
        self._dbstmnt.bind_in("src", self.irec.src, str, 128)
        self._dbstmnt.bind_in("tm_stamp", self.irec.tm_stamp, datetime.datetime)

        await self._dbcon.execute(self._dbstmnt)

        self.irec.id = await self._dbcon.post_get_sequence(self._dbstmnt, "audit.Tran", "id", 8)

        if irec:
            irec._copy_from(self.irec)

        return self

